/* Copyright (C) 2010 Stephane Rochoy
 * 
 * This file is part of Weapon Factory.
 * 
 * Weapon Factory is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Weapon Factory is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Weapon Factory. If not, see <http://www.gnu.org/licenses/>.
 */
#include "Engine.hh"

#include "GameInputListener.hh"
#include "Core.hh"

using namespace WeaponFactory;

Engine::Engine()
  : mFrameListener(0),
    mGameInputListener(0),
    mCore(0),
    mInputDispatcher(0),
    mInputManager(0),
    mKeyboard(0),
    mMouse(0),
    mResourcePath("/home/sheda/devel/weapon-factory/resources/"), 
    mRoot(0)
{
}

Engine::~Engine()
{
  Ogre::LogManager::getSingletonPtr()
    ->logMessage("*** Destroying game ***");
  if (mGameInputListener) {
    delete mGameInputListener;
    mGameInputListener = 0;
  }
  if (mInputDispatcher) {
    delete mInputDispatcher;
    mInputDispatcher = 0;
  }
  if (mInputManager) {
    if (mKeyboard) {
      mInputManager->destroyInputObject(mKeyboard);
      mKeyboard = 0;
    }
    if (mMouse) {
      mInputManager->destroyInputObject(mMouse);
      mMouse = 0;
    }
    OIS::InputManager::destroyInputSystem(mInputManager);
    mInputManager = 0;
  }
  if (mWindow) {
    /** Detaching OIS before window shutdown is done through InputListener's
     *  destructor. Very important under GNU/Linux.
     */
    //delete mWindow; /// \todo Why is it core dumping?
    mWindow = 0;
  }
  if (mFrameListener) {
    delete mFrameListener;
    mFrameListener = 0;
  }
  if (mRoot) {
    delete mRoot;
    mRoot = 0;
  }
}

void 
Engine::run(void)
{
  if (!setup()) {
    return;
    // NOTREACHED
  }
  Ogre::LogManager::getSingletonPtr()
    ->logMessage("~~~~~~~~~~ ENTERING GAME ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
  mCore->setRunning(true); // Enter the game.
  mRoot->startRendering();    
  Ogre::LogManager::getSingletonPtr()
    ->logMessage("~~~~~~~~~~ GAME OVER ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
  destroyScene();
}

bool
Engine::setup(void)
{
  using namespace Ogre;

  mRoot = new Root(mResourcePath + "plugins.cfg", 
                   mResourcePath + "ogre.cfg", 
                   mResourcePath + "ogre.log");
  LogManager::getSingletonPtr()
    ->logMessage("*** Loading game ***");
  setupResources();
  if (!configure()) {
    return false;
    // NOTREACHED
  }
  createWindow();
  chooseSceneManager();
  createSceneNodes();
  createCamera();
  createViewports();
  loadResources();
  createCore();
  createUserInterface();
  createInputSystem();
  createFrameListener();
  return true;
}

void
Engine::setupResources(void)
{
  using namespace Ogre;

  ConfigFile cf;

  cf.load(mResourcePath + "resources.cfg");
  ConfigFile::SectionIterator seci = cf.getSectionIterator();
  while (seci.hasMoreElements()) {
    String secName = seci.peekNextKey();
    ConfigFile::SettingsMultiMap* settings = seci.getNext();
    ConfigFile::SettingsMultiMap::iterator i;
    for (i = settings->begin(); i != settings->end(); ++i) {
      String typeName = i->first;
      String archName = i->second;
      ResourceGroupManager::getSingleton()			\
	.addResourceLocation(archName, typeName, secName);
    }
  }
}

void
Engine::loadResources(void)
{
  using namespace Ogre;

  // Set default mipmap level (NB some APIs ignore this).
  TextureManager::getSingleton().setDefaultNumMipmaps(5);
  ResourceGroupManager::getSingleton().initialiseAllResourceGroups();
}

/** Try to restore the last configuration if any. In case of failure, prompt the
 *  user for configuration. Return false if both no configuration can be
 *  restored and the user cancelled the configuration dialog box.
 */
bool
Engine::configure(void)
{
  return mRoot->restoreConfig() || mRoot->showConfigDialog();
}

void
Engine::createWindow(void)
{
  using namespace Ogre;

  LogManager::getSingletonPtr()
    ->logMessage("Creating window...");
  mWindow = mRoot->initialise(true);
  WindowEventUtilities::addWindowEventListener(mWindow, this);
}

void
Engine::chooseSceneManager(void)
{
  using namespace Ogre;

  mSceneMgr = mRoot->createSceneManager(ST_GENERIC, "main.scene-manager.weapon-factory");
}

void
Engine::createSceneNodes(void)
{
  using namespace Ogre;

  LogManager::getSingletonPtr()
    ->logMessage("Creating scene nodes...");
  mSceneMgr
    ->getRootSceneNode()
    ->createChildSceneNode("main-camera.node.weapon-factory",
                           Vector3(0, 500, 500));
  mSceneMgr
    ->getRootSceneNode()
    ->createChildSceneNode("trooper.node.weapon-factory", 
                           Vector3(0, 0, -50));
}

void
Engine::createCamera(void) 
{
  using namespace Ogre;

  LogManager::getSingletonPtr()
    ->logMessage("Creating camera...");

  mCamera = mSceneMgr->createCamera("main-camera.camera.weapon-factory");
  mCamera->setNearClipDistance(5);
  SceneNode* cameraNode 
    = mSceneMgr->getSceneNode("main-camera.node.weapon-factory");
  cameraNode->attachObject(mCamera);
  mCamera->setPosition(Vector3(0,0,0));

  SceneNode* node = mSceneMgr->getSceneNode("trooper.node.weapon-factory");
  mCamera->lookAt(node->getPosition());
  mCamera->setFixedYawAxis(true, node->getLocalAxes().GetColumn(1));
}

void
Engine::createViewports(void) 
{
  using namespace Ogre;

  LogManager::getSingletonPtr()
    ->logMessage("Creating viewports...");

  Viewport* vp;
  vp = mWindow->addViewport(mCamera);
  vp->setBackgroundColour(ColourValue(0, 0, 0));
  mCamera->setAspectRatio(Real(vp->getActualWidth())
                          / Real(vp->getActualHeight()));
}

void
Engine::createGround(void) 
{
  using namespace Ogre;

  MeshManager::getSingleton()
    .createPlane("ground.plane.weapon-factory",
	         ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME, 
	         Plane(Vector3::UNIT_Y, 0),
	         1500, 1500, 20, 20, true, 1, 5, 5, Vector3::UNIT_Z);
  Ogre::Entity *plane 
    = mSceneMgr->createEntity("ground.entity.weapon-factory",
                              "ground.plane.weapon-factory");
  mSceneMgr
    ->getRootSceneNode()
    ->createChildSceneNode()
    ->attachObject(plane);
  plane->setMaterialName("WeaponFactory/Ground");
  plane->setCastShadows(false);
}

void
Engine::createEntities(void) 
{
  mCore->createEntity("robot", "trooper");
}

void
Engine::createUserInterface(void)
{
  mUserInterface = new UserInterface(mWindow, mSceneMgr, mCore);
}

void
Engine::createCore(void)
{
  using namespace Ogre;

  LogManager::getSingletonPtr()
    ->logMessage("Creating scene...");
  mSceneMgr->setAmbientLight(ColourValue(1, 1, 1));
  createGround();
  mCore = new Core(mCamera->getParentSceneNode());
  createEntities();
}

/** Create the input system.
 *  Create the OIS::Keyboard and OIS::Mouse input objects used by the
 *  InputDispatcher. Create the InputDispatcher itself as well as its
 *  InputListeners.
 */ 
void
Engine::createInputSystem(void) 
{
  using namespace Ogre;

  LogManager::getSingletonPtr() \
    ->logMessage("*** Initializing input system ***");

  // Create the OIS::Keyboard and OIS::Mouse input objects.
  size_t             windowHnd = 0;
  std::ostringstream windowHndStr;
  OIS::ParamList     pl;
  mWindow->getCustomAttribute("WINDOW", &windowHnd);
  windowHndStr << windowHnd;
  pl.insert(std::make_pair(std::string("WINDOW"), windowHndStr.str()));
  pl.insert(std::make_pair(std::string("XAutoRepeatOn"), std::string("true")));
  mInputManager = OIS::InputManager::createInputSystem(pl);
  mKeyboard = static_cast<OIS::Keyboard*>(mInputManager
    ->createInputObject(OIS::OISKeyboard, true));
  mMouse = static_cast<OIS::Mouse*>(mInputManager
      ->createInputObject(OIS::OISMouse, true));
  mMouse->getMouseState().height = mWindow->getHeight();
  mMouse->getMouseState().width  = mWindow->getWidth();

  // Create the InputDispatcher and its InputListeners.
  mInputDispatcher = new InputDispatcher(mKeyboard, mMouse);
  mGameInputListener = new GameInputListener(mResourcePath + "key-bindings.list", 
                                             mCore);
  mInputDispatcher->bindListener(mGameInputListener, "play");
  mInputDispatcher->pushContext("play");
}

void
Engine::createFrameListener(void) 
{
  mFrameListener = new FrameListener(mInputDispatcher, mCore);
  mRoot->addFrameListener(mFrameListener);
}

/** Destroy both the scene and the Core.
 */
void
Engine::destroyScene(void) 
{
  using namespace Ogre;
  
  LogManager::getSingletonPtr()
    ->logMessage("Destroying scene...");
  LogManager::getSingletonPtr()
    ->logMessage("Destroying game state...");
  delete mCore;
  mCore = 0;
}

// Ogre::WindowEventListener interface ////////////////////////////////////////

bool
Engine::windowClosing(Ogre::RenderWindow* window) 
{
  if (window == mWindow) {
    Ogre::LogManager::getSingletonPtr()
      ->logMessage("Window is closing...");
  }
}
