/* Copyright (C) 2010 Stephane Rochoy
 * 
 * This file is part of Weapon Factory.
 * 
 * Weapon Factory is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Weapon Factory is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Weapon Factory. If not, see <http://www.gnu.org/licenses/>.
 */
#include "InputListener.hh"

#include <fstream>

#include "Exception.hh"
#include "Ogre.h"

using namespace WeaponFactory;

InputListener::InputListener(void* opaqueData)
  : mOpaqueData(opaqueData) 
{
}

void
InputListener::bindKey(OIS::KeyCode key, const std::string actionName)
{
  std::map<const std::string, Functor*>::iterator k
    = mActions.find(actionName);
  if (k == mActions.end()) {
    // Unable to find actionName in mActions.
    Ogre::LogManager::getSingletonPtr()
      ->logMessage("Unknon action " + actionName);
    throw Exception();
  }
  mKeyBindings[key] = mActions[actionName];
}

void
InputListener::loadKeyBindings(const std::string fname)
{
  using namespace Ogre;
  using namespace std;

  LogManager::getSingletonPtr()
    ->logMessage("Loading key bindings from " + fname);
  ifstream ifs(fname.c_str());
  if (ifs.is_open()) {
    const unsigned int buffer_size = 256;
    char               buffer[buffer_size];
    
    for (unsigned int lineNum = 1; ifs.good(); lineNum++) {
      ifs.getline(buffer, buffer_size);
      if (ifs.gcount() == buffer_size) {
	// Buffer is too small to store the line.
	LogManager::getSingletonPtr()
	  ->logMessage("Line is too big"); /// \todo Find a way to log lineNum here.
        throw Exception();
      }
      string line(buffer);
      if (line.size() == 0) {
        continue;
      }
      unsigned int pos = line.find_first_of(" \t");
      string actionName = line.substr(0, pos);
      pos = line.find_first_not_of(" \t", pos);
      string keySeq = line.substr(pos);
      if (keySeq == "esc") {
        this->bindKey(OIS::KC_ESCAPE, actionName);
      } else if (keySeq == "arrow-down") {
        this->bindKey(OIS::KC_DOWN, actionName);
      } else if (keySeq == "arrow-left") {
        this->bindKey(OIS::KC_LEFT, actionName);
      } else if (keySeq == "arrow-right") {
        this->bindKey(OIS::KC_RIGHT, actionName);
      } else if (keySeq == "arrow-up") {
        this->bindKey(OIS::KC_UP, actionName);
      } else if (keySeq == "f1") {
        this->bindKey(OIS::KC_F1, actionName);
      } else if (keySeq == "f2") {
        this->bindKey(OIS::KC_F2, actionName);
      } else if (keySeq == "space") {
        this->bindKey(OIS::KC_SPACE, actionName);
      } else {
        LogManager::getSingletonPtr()
          ->logMessage("Unknown key" + keySeq);
        continue;
      }
      LogManager::getSingletonPtr()
        ->logMessage("Action " + actionName
                     + " bound to key " + keySeq);
    }
  } else {
    // Unable to open file.
    LogManager::getSingletonPtr()
      ->logMessage("Unable to open file " + fname);
    throw Exception();
  }
}

/*
void
InputListener::registerAction(const std::string name, Action handler)
{
  mActions[name] = handler;
  Ogre::LogManager::getSingletonPtr()
    ->logMessage("Action " + name + " registered");
}
*/

void
InputListener::registerAction(const std::string name, Functor* functor)
{
  mActions[name] = functor;
  Ogre::LogManager::getSingletonPtr()
    ->logMessage("Action " + name + " registered");
}

void
InputListener::triggerAction(const std::string name)
{
  //Action action = mActions[name];
  Functor* functor = mActions[name];
  (*functor)();
}

// OIS::KeyListener interface /////////////////////////////////////////////////

bool
InputListener::keyPressed(const OIS::KeyEvent &e) 
{
  // Inject keys into CEGUI if no builtin action was triggered.
  CEGUI::System &sys = CEGUI::System::getSingleton();
  if (sys.injectKeyDown(e.key)
      || sys.injectChar(e.text)) {
    return true;
  }

  // Handle key via key bindings if not processed by CEGUI.
  if (mKeyBindings.find(e.key) != mKeyBindings.end()) {
    // The key was found in the key bindings.
    //Action action = mKeyBindings[e.key];
    Functor* functor = mKeyBindings[e.key];
    (*functor)();
    return true;
  }
  return true;
}

bool 
InputListener::keyReleased(const OIS::KeyEvent &e) {
  if (CEGUI::System::getSingleton().injectKeyUp(e.key)) {
    return true;
  }
  return true;
}

// OIS::MouseListener interface ///////////////////////////////////////////////

bool 
InputListener::mouseMoved(const OIS::MouseEvent &e)
{
  CEGUI::System& system = CEGUI::System::getSingleton();
  if (system.injectMouseMove(e.state.X.rel, e.state.Y.rel)) {
  }
  if (e.state.Z.rel) {
    // Scroll the wheel.
    system.injectMouseWheelChange(e.state.Z.rel / 120.0f);
  }
  return true;
}

bool 
InputListener::mousePressed(const OIS::MouseEvent &e, OIS::MouseButtonID id)
{
  if (CEGUI::System::getSingleton()
    .injectMouseButtonDown(convertMouseButton(id))) {
    return true;
  }
  return true;
}

bool 
InputListener::mouseReleased(const OIS::MouseEvent &e, OIS::MouseButtonID id)
{
  if (CEGUI::System::getSingleton()
    .injectMouseButtonUp(convertMouseButton(id))) {
    return true;
  }
  return true;
}

// Utility /////////////////////////////////////////////////////////////////////

CEGUI::MouseButton 
InputListener::convertMouseButton(OIS::MouseButtonID buttonID)
{
    switch (buttonID) {
    case OIS::MB_Left:
        return CEGUI::LeftButton;
    case OIS::MB_Right:
        return CEGUI::RightButton;
    case OIS::MB_Middle:
        return CEGUI::MiddleButton;
    default:
        return CEGUI::LeftButton;
    }
}
